package sk.tuke.bkpi.generator;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import sk.tuke.bkpi.model.Module;
import sk.tuke.bkpi.model.Project;
import sk.tuke.bkpi.model.clause.ComponentReference;
import sk.tuke.bkpi.model.component.Implementation;
import sk.tuke.bkpi.model.component.Machine;
import sk.tuke.bkpi.model.data.DataField;
import sk.tuke.bkpi.model.data.ForeignVariable;

/**
 * Manages mapping of B-Method modules data and composition links to their names
 * in target language code. It use the same name of data or link for target language
 * when it's possible. But if there is a conflict in using its name in target language,
 * new name will be generated. It supports multiple visibility scopes and target
 * language keywords definition.
 *  
 * @author Marcel Dancak
 */
public class NameManager {

	/**
	 * Data table for all data with module visibility level.
	 * 
	 * @param <T> The type of the module data
	 */
	static class DataTable<T> {
		
		private Stack<Map<T, String>> scopes = new Stack<Map<T,String>>();
		
		public DataTable() {
			openScope();
		}
		
		public Map<T, String> scopeTable() {
			return scopes.peek();
		}
		
		public void openScope() {
			scopes.push(new Hashtable<T, String>());
		}
		
		public void closeScope() {
			if (scopes.isEmpty()) {
				throw new IllegalStateException("Cannot close global scope");
			}
			scopes.pop();
		}
		
		public void addData(T data, String name) {
			if (scopeTable().containsValue(name)) {
				throw new IllegalArgumentException("Current scope already contains name: "+name);
			}
			scopeTable().put(data, name);
		}
		
		public boolean containsData(T data) {
			Iterator<Map<T, String>> it = scopes.iterator();
			while(it.hasNext()) {
				if (it.next().containsKey(data))
					return true;
			}
			return false;
		}
		
		public boolean containsGlobalData(String name) {
			return scopes.get(0).containsValue(name);
		}
		
		public boolean containsLocalData(String name) {
			Iterator<Map<T, String>> it = scopes.iterator();
			it.next(); //skip global scope
			
			while(it.hasNext()) {
				Map<T, String> scopeData = it.next();
				if (scopeData.containsValue(name)) {
					return true;
				}
			}
			return false;
		}
		
		public boolean containsName(String name) {
			Iterator<Map<T, String>> it = scopes.iterator();
			
			while(it.hasNext()) {
				Map<T, String> scopeData = it.next();
				if (scopeData.containsValue(name)) {
					return true;
				}
			}
			
			return false;
		}
		
		public String nameOf(T data) {
			Iterator<Map<T, String>> it = scopes.iterator();
			while(it.hasNext()) {
				Map<T, String> scopeData = it.next();
				if (scopeData.containsKey(data))
					return scopeData.get(data);
			}
			return null;
		}
		
		public void dumpInfo() {
			System.out.println("+++++++++++++++++++");
			for (Map<T, String> scope : scopes) {
				System.out.println(scope);
			}
			System.out.println("+++++++++++++++++++");
		}
	}
	
	
	private Collection<String> keywords = Collections.emptyList();
	
	private Module workingModule;
	private Map<Module, DataTable<DataField>> modulesDataTable = new Hashtable<Module, DataTable<DataField>>();
	private Map<Module, Map<ComponentReference, String>> modulesChainsTable = new Hashtable<Module, Map<ComponentReference,String>>();
	
	/*Atribute of Ivan Halupka*/
	private MachineNameManager machine_name_manager = new MachineNameManager();
	private OperationsNameManager operations_name_manager = new OperationsNameManager();
	private RecordItemsNameManager record_items_name_manager = new RecordItemsNameManager();
	
	public void SetRecordExcludedWords ()
	{
		record_items_name_manager.setExcluded_words(this.keywords);	
	}
	
	public Map<String, List<String>> getImplementations_per_machine()
	{
		return this.operations_name_manager.getImplementations_per_machine();
	}
	
	public String record_data (String record_class_name, String record_item_name)
	{
		return this.record_items_name_manager.record_data(record_class_name, record_item_name);
	}
	
	public void generateUniqueMachineNames (Project project)
	{
		this.machine_name_manager.setExcluded_words(this.keywords);
		this.machine_name_manager.generate_machine_names(project); 
	}
	
	public void resolveOperationNames(Project project)
	{
		this.operations_name_manager.setExcluded_words(this.keywords);
		this.operations_name_manager.generate_operation_names(project);
	}
	
    public String generateUniqueClassName (String proposal)
    {
    	return this.machine_name_manager.generateUniqueClassName(proposal);
    }
	
    public String generateMainMachineName()
    {
    	return this.machine_name_manager.generateMainMachineName();
    }
	
	/**
	 * Sets keywords of the target language.
	 * @param keywords The list of keywords
	 */
	public void setKeywords(String ... keywords) {
		this.keywords = Arrays.asList(keywords);
	}
	
	/**
	 * Sets keywords of the target language.
	 * @param keywords The list of keywords
	 */
	public void setKeywords(Collection<String> keywords) {
		this.keywords = keywords;
	}
	
	private DataTable<DataField> moduleTable(Module module) {
		if (!modulesDataTable.containsKey(module)) {
			DataTable<DataField> dataTable = new DataTable<DataField>();
			modulesDataTable.put(module, dataTable);
			return dataTable;
		}
		return modulesDataTable.get(module);
	}
	
	public String componentReferenceName(ComponentReference chain) {
		return componentReferenceName(workingModule, chain);
	}
	
	public String componentReferenceName(Module module, ComponentReference chain) {
		if (!modulesChainsTable.containsKey(module)) {
			modulesChainsTable.put(module, new Hashtable<ComponentReference, String>());
		}
		Map<ComponentReference, String> chainsTable = modulesChainsTable.get(module);
		if (chain == null)
			throw new NullPointerException();
		if (chainsTable.containsKey(chain)) {
			return chainsTable.get(chain);
		}
		String name = chain.prefixString()!=null && chain.prefixString().length() > 0? chain.prefixString().replace(".", "_") : chain.machine().name().toLowerCase();
		DataTable<DataField> dataTable = moduleTable(module);
		if (keywords.contains(name) || dataTable.containsName(name)) { /*should be modified to while - I think*/
			name = name+"_ref";
		}
		chainsTable.put(chain, name);
		return name;
	}
	
	/**
	 * Sets working module for more convenient use of the other methods,
	 * without this module as argument. 
	 * @param module
	 */
	public void setWorkingModule(Module module) {
		workingModule = module;
	}
	
	/*
	boolean contains(Module module, String name) {
		return moduleTable(module).containsName(name) || keywords.contains(name);
	}*/
	
	/**
	 * Returns name for target language of the <code>data</code> for working module.
	 */
	public String moduleData(DataField data) {
		return moduleData(workingModule, data);
	}
	
	/**
	 * Returns name for target language of the <code>data</code> for <code>module</code>.
	 */
	public String moduleData(Module module, DataField data) {
		/*
		if (data instanceof ForeignData) {
			return moduleData(module, ((ForeignData) data).data());
		}
		*/
		DataTable<DataField> dataTable = moduleTable(module);
		
		if (dataTable.containsData(data)) {
			return dataTable.nameOf(data);
		}
		
		String name = data.name();
		while(true) {
			// allow one local variable with the same name as global data
			if (dataTable.containsGlobalData(name) && !dataTable.containsLocalData(name) && !keywords.contains(name))
				;//break;
			
			if (!dataTable.containsName(name) && !keywords.contains(name))
				break;
			dataTable.dumpInfo();
			System.out.println("doesn contains "+data);
			name = name + "_data";
		}
	
		dataTable.addData(data, name);
		return name;
	}
	
	public String globalModuleData(Module module, DataField data) {
		DataTable<DataField> dataTable = moduleTable(module);
		
		for (DataField d : dataTable.scopes.get(0).keySet()) {
			if (d.name().equals(data.name()))
				return dataTable.scopes.get(0).get(d);
		}
		throw new IllegalArgumentException("Module "+module+" doesn't contain data "+data);
	}
	//
	
	//public String componentReferenceName(ComponentReference chain)
	//public String componentReferenceName(Module module, ComponentReference chain)
	//public String moduleData(DataField data)
	//public String moduleData(Module module, DataField data)
	public String generateUniqueName(String proposal) {
		return generateUniqueName(workingModule, proposal);
	}
	
	public String generateUniqueName(Module module, String proposal) {
		//String name =  names.peek().contains(base)? "_"+base : base;
		//String name = containsName(base)? "_"+base : base;
		String name = proposal;
		DataTable<DataField> dataTable = moduleTable(module);
		while(dataTable.containsName(name) || keywords.contains(name)) {
			name = "_"+name;
		}
		return name;
	}
	
	
	/*Method of Ivan Halupka*/
	public String getUniqueMachineName()
	{
		String unique_machine_name = this.machine_name_manager.getResolvedName(this.workingModule.machine().name());
		return unique_machine_name;
	}
	
	/*Method of Ivan Halupka*/	
	public String getUniqueMachineName (String machine)
	{
		String unique_machine_name = this.machine_name_manager.getResolvedName(machine);
		return unique_machine_name;		
	}
	
	public String get_resolved_operation_name(Machine machine, String operation_name)
	{
		return this.operations_name_manager.get_resolved_operation_name(machine, operation_name);
	}
	
	public String get_resolved_operation_name(Implementation implementation, String operation_name)
	{
		return this.operations_name_manager.get_resolved_operation_name(implementation, operation_name);
	}
	
	public String generate_unique_method_name (Machine machine, String proposal)
	{
		return this.operations_name_manager.generate_unique_method_name(machine, proposal);
	}

	public String generate_unique_method_name (Implementation implementation, String proposal)
	{
		return this.operations_name_manager.generate_unique_method_name(implementation, proposal);	
	}

	
	
	
	public void openScope() {
		openScope(workingModule); 
	}
	
	public void openScope(Module module) {
		moduleTable(module).openScope();
	}
	
	public void closeScope() {
		closeScope(workingModule);
	}
	
	public void closeScope(Module module) {
		moduleTable(module).closeScope();
	}
}
