package sk.tuke.bkpi.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Vector;
import javax.swing.ComboBoxModel;
import javax.swing.DefaultComboBoxModel;
import javax.swing.DefaultListModel;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.ListModel;
import javax.swing.UIManager;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;

import javax.tools.Diagnostic;
import javax.tools.DiagnosticCollector;
import javax.tools.JavaCompiler;
import javax.tools.JavaFileObject;
import javax.tools.StandardJavaFileManager;
import javax.tools.StandardLocation;

import org.apache.log4j.PropertyConfigurator;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import sk.tuke.bkpi.extensions.CompiledImplementationManager;
import sk.tuke.bkpi.extensions.ExcludedWordsManager;
import sk.tuke.bkpi.generator.CsGenerator;
import sk.tuke.bkpi.generator.JavaCodeGenerationException;
import sk.tuke.bkpi.generator.JavaGenerator;
import sk.tuke.bkpi.model.Module;
import sk.tuke.bkpi.model.Project;
import sk.tuke.bkpi.parser.ProjectParseException;
import sk.tuke.bkpi.parser.ProjectParser;

/**
 * Provides simple graphic user interface for BKPI compiler with
 * several additional function. More precisely, it provide:
 * <ul>
 *  <li> compilation of B-Method projects to Java programming language</li>
 *  <li> compilation of generated Java code to bytecode by javac </li>
 *  <li> runs compiled project - Java bytecode </li>
 *  <li> shows project graph image </li>
 *  <li> keeps list of recent working projects in file </li>
 * </ul>
 * 
 * @author Marcel Dancak
 */
public class SimpleGui extends javax.swing.JFrame {

	private static final long serialVersionUID = 1L;
	private static String OS = System.getProperty("os.name");

	public static final String NO_PROJECT_ERROR_MESSAGE = "No project selected!";
	public static final String NO_JAVAC_ERROR_MESSAGE = "Can't find Java compiler!";
	public static final String NO_GRAPHVIZ_ERROR_MESSAGE = "Graphviz library not installed!";
	public static final File GRAPH_TEMPLATE = new File("freemarker.template");

	public static final String RECENT_FILE = "recent";
	static {
		sk.tuke.bkpi.logging.Logger.configure("log4j.properties");
	}

	private static Logger logger = sk.tuke.bkpi.logging.Logger
			.getLogger(SimpleGui.class);

	private Vector<String> recentProjects;
	private HtmlConsole console;

	private Project bProject;
	private File javaDir;
	private File classDir;
	private File libraryDir = new File("BasicMachines/BKpiBasic.jar");

	private OutputStream stdin;

	public SimpleGui() {
		loadRecentProjectsList();
		initComponents();
		console = new HtmlConsole();
		consoleScrollPane.setViewportView(console);

		projectSelected(null);

		HtmlConsoleAppender.htmlConsole = console;

		console.addHyperlinkListener(new HyperlinkListener() {

			public void hyperlinkUpdate(HyperlinkEvent e) {
				if (e.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
					System.out.println(e.getURL());
					openFile(e.getURL().toString());
				}
			}
		});
	}

	/**
	 * Try to open file <code>path</code> in an external editor.
	 * @param path name of the file
	 */
	private void openFile(String path) {
		try {
			String editor;
			if (OS.equals("Linux"))
				editor = "gedit";
			else
				editor = "notepad";
			Process p = Runtime.getRuntime().exec(editor + " " + path);
		} catch (IOException e1) {
			e1.printStackTrace();
		}
	}

	public ComboBoxModel recentDirectories() {
		return new DefaultComboBoxModel(recentProjects);
	}

	/**
	 * Reads a list of recently used projects from file if exist.
	 */
	private void loadRecentProjectsList() {
		File file = new File(RECENT_FILE);
		if (file.exists()) {
			try {
				FileReader fr = new FileReader(file);
				BufferedReader br = new BufferedReader(fr);

				recentProjects = new Vector<String>();
				String line = br.readLine();
				while (line != null) {
					recentProjects.add(line);
					line = br.readLine();
				}
			} catch (FileNotFoundException ex) {
				logger.log(Level.ERROR, null, ex);
			} catch (IOException ex) {
				logger.log(Level.ERROR, null, ex);
			}
		} else {
			recentProjects = new Vector<String>();
		}
	}

	/**
	 * Saves a list of recently used projects to the file.
	 */
	private void saveRecentProjectsList() {
		File file = new File(RECENT_FILE);
		if (!file.exists()) {
			try {
				file.createNewFile();
			} catch (IOException ex) {
				logger.log(Level.ERROR, null, ex);
			}
		}
		OutputStreamWriter outputStream = null;
		try {
			outputStream = new FileWriter(file);
			PrintWriter writer = new PrintWriter(outputStream);
			for (String elem : recentProjects) {
				writer.println(elem);
			}
			writer.flush();
		} catch (IOException ex) {
			logger.log(Level.ERROR, null, ex);
		} finally {
			if (outputStream != null)
				try {
					outputStream.close();
				} catch (IOException ex) {
					logger.log(Level.ERROR, null, ex);
				}
		}
	}

	/**
	 * Builds a list (model for graphic components) of all B-Method project source files
	 * for specified directory <code>dir</code>.
	 * 
	 * @param dir project directory
	 * @return list of source files as <code>ListModel</code>
	 */
	private ListModel model(File dir) {
		DefaultListModel model = new DefaultListModel();
		for (File file : collectProjectFiles(dir))
			model.addElement(new FileItem((file)));
		return model;
	}

	/**
	 * Lists files inside directory <code>dir</code> and filter them
	 * by their extensions.
	 * 
	 * @param dir project directory
	 * @return list of B-language components files
	 */
	private List<File> collectProjectFiles(File dir) {
		List<File> list = new ArrayList<File>();
		listDirectory(dir, list);
		return list;
	}

	private void listDirectory(File dir, List<File> list) {
		for (File f : dir.listFiles()) {
			if (f.isDirectory())
				;//listDirectory(f, list); // bez rekurzie, AtelierB tam ma kopie
			else {
				if (f.getName().endsWith(".mch")
						|| f.getName().endsWith(".ref")
						|| f.getName().endsWith(".imp"))
					list.add(f);
			}
		}
	}

	private File projectFolder() {
		Object selected = jComboBox1.getSelectedItem();
		if (selected == null)
			return null;
		File projectFolder = new File(jComboBox1.getSelectedItem().toString());
		return projectFolder;
	}

	/**
	 * Parses selected project files and generate Java code. Data model of the project
	 * will stay stored in data field <code>bProject</code> and will be available for another use.
	 */
	private void parseProject() {

		final File projectFolder = projectFolder();
		if (projectFolder == null) {
			showError(NO_PROJECT_ERROR_MESSAGE);
			return;
		}

		long ts = System.currentTimeMillis();
		StringWriter buffer = new StringWriter();
		ProjectParser parser = new ProjectParser(projectFolder);

		try {
			parser.setLibraries("BasicMachines/MCH");
			bProject = parser.parseProject(compiledImplementationsManager
					.AllMachineImplementationPairs("Java"));

			/*
			for (Module m: bProject.modules())
			{
				System.out.println("BLACH : "+ m.implementation().name());
			}*/

		} catch (ProjectParseException e) {
			logger.error("Error in parsing file:");
			logger.error(e.getSource());
			logger.error(e.getMessage());
			e.printStackTrace();
			return;
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return;
		}
		File javaOutputFolder = new File(projectFolder, "java");

		JavaGenerator javaGenerator = new JavaGenerator(bProject,
				javaOutputFolder);
		javaGenerator.resetJavaExcludedWords();
		javaGenerator.setJavaExcludedWords(EWM.getExcludedWords("Java"));

		javaGenerator.setModificationPolicy(IMPM.getModificationPolicyManager()
				.getLanguageModificationPolicy("Java"));

		javaGenerator.setAccessPolicy(IAPM.getAccessPolicyManager()
				.getAccessPolicy("Java"));

		//MIPAIRS

		/*ADDED BY IVAN BEGIN */

		try {
			parser.setLibraries("BasicMachines/MCH");
			bProject = parser.parseProject(compiledImplementationsManager
					.AllMachineImplementationPairs("Csharp"));

			/*
			for (Module m: bProject.modules())
			{
				System.out.println("BLACH : "+ m.implementation().name());
			}*/

		} catch (ProjectParseException e) {
			logger.error("Error in parsing file:");
			logger.error(e.getSource());
			logger.error(e.getMessage());
			e.printStackTrace();
			return;
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return;
		}

		File csOutputFolder = new File(projectFolder, "c#");
		CsGenerator csGenerator = new CsGenerator(bProject, csOutputFolder);
		csGenerator.ResetCsExcludedWords();
		csGenerator.setCsExcludedWords(EWM.getExcludedWords("Csharp"));
		csGenerator.setModificationPolicy(IMPM.getModificationPolicyManager()
				.getLanguageModificationPolicy("Csharp"));
		csGenerator.setAccessPolicy(IAPM.getAccessPolicyManager()
				.getAccessPolicy("Csharp"));

		/*ADDED BY IVAN END*/

		logger.info("\nGenerating Java Code ...");
		try {
			javaGenerator.generateJavaCode(buffer);
			csGenerator.generateCsCode(buffer);

		} catch (JavaCodeGenerationException jcge) {
			logger.error("Error in generating java code for module: "
					+ jcge.getSource());
			logger.error(jcge.getMessage());
			jcge.printStackTrace();
			return;
		}
		long te = System.currentTimeMillis();
		System.out.println(te - ts);
		//log.info("compilation time: "+(te-ts));
		for (Module m : bProject.modules()) {
			System.out.println(m.machine() + " " + m.refinements() + " "
					+ m.implementation());
		}
	}

	/**
	 * Parse selected project files and generate Java code in a new thread.
	 */
	private void parseProjectAsynchronously() {
		//log.trace("this is trace");
		//log.debug("this is debug");
		//log.warn("this is warrning");

		Runnable task = new Runnable() {
			public void run() {
				parseProject();
			}
		};
		new Thread(task).start();
	}

	private void showGraphAsynchronously() {
		Runnable task = new Runnable() {
			public void run() {
				showGraph();
			}
		};
		new Thread(task).start();
	}

	/**
	 * Shows dialog window with graph image of the selected project.
	 * If the project isn't parsed, it will parse it first.
	 */
	private void showGraph() {
		if (projectFolder() == null) {
			showError(NO_PROJECT_ERROR_MESSAGE);
			return;
		}
		try {
			Process p = Runtime.getRuntime().exec("dot -V");
			if (p.waitFor() != 0)
				showError(NO_GRAPHVIZ_ERROR_MESSAGE);
		} catch (Exception ex) {
			showError(NO_GRAPHVIZ_ERROR_MESSAGE);
		}

		File graphFile = new File(projectFolder(), "graph.png");
		// && !graphFile.exists()
		if (bProject == null) {
			parseProject();
			if (bProject == null) // not successful parsing
				return;

		}
		GraphvizVisualization graphGenerator = new GraphvizVisualization(
				bProject, GRAPH_TEMPLATE, graphFile);

		ImagePanel graphImage = null;
		ImagePanel legend = null;
		try {
			graphGenerator.generateGraph();

			graphImage = new ImagePanel(new File(projectFolder(), "graph.png"));
			//graphImage = new ImagePanel(new File(projectFolder() + "\\" + "graph.png"));

			legend = new ImagePanel("graph_legend.png");

		} catch (Exception ex) {
			logger.log(Level.ERROR,
					"error during generating graph: " + ex.getMessage());
		}

		graphImage.setBackground(Color.white);
		legend.setOpaque(false);
		legend.setBackground(new Color(0, 0, 0, 0));
		legend.setSize(legend.getPreferredSize());
		// legend.setLocation(10, 10);
		graphImage.add(legend);

		JPanel graph = new JPanel();
		graph.setLayout(new GridBagLayout());
		graph.setBackground(Color.white);
		GridBagConstraints c = new GridBagConstraints();
		c.weightx = 1.0;
		c.gridx = 0;
		c.gridy = 0;
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.NONE;
		graph.add(legend, c);

		c.gridy = 1;
		c.anchor = GridBagConstraints.NORTH;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.weighty = 1.0;
		graph.add(graphImage, c);

		jScrollPane3.setViewportView(graph);
		JDialog dialog = new JDialog(this, "Project Graph");

		dialog.add(jScrollPane3, BorderLayout.CENTER);

		dialog.pack();
		dialog.setLocation(6, 30);
		dialog.setVisible(true);
	}

	/**
	 * Shows error dialog window with message <code>text</code>
	 * @param text
	 */
	private void showError(String text) {
		JOptionPane.showMessageDialog(this, text, "Error",
				JOptionPane.ERROR_MESSAGE);
	}

	private void compileToJava() {
		console.clear();
		File projectFile = projectFolder();
		if (projectFile == null) {
			showError(NO_PROJECT_ERROR_MESSAGE);
			return;
		}
		System.out.println(projectFile);

		if (!classDir.exists()) {
			classDir.mkdir();
		}
		JavaCompiler compiler = javax.tools.ToolProvider
				.getSystemJavaCompiler();
		if (compiler == null) {
			showError(NO_JAVAC_ERROR_MESSAGE);
			return;
		}

		DiagnosticCollector<JavaFileObject> diagnostics = new DiagnosticCollector<JavaFileObject>();
		StandardJavaFileManager fileManager = compiler.getStandardFileManager(
				diagnostics, null, null);

		try {
			fileManager.setLocation(StandardLocation.CLASS_OUTPUT,
					Collections.singleton(classDir));
			fileManager.setLocation(StandardLocation.CLASS_PATH,
					Collections.singleton(libraryDir));
		} catch (IOException ex) {
			logger.log(Level.ERROR, null, ex);
		}

		Iterable<? extends JavaFileObject> compilationUnits = fileManager
				.getJavaFileObjects(javaDir.listFiles(new FileFilter() {

					public boolean accept(File file) {
						return file.getName().endsWith(".java");
					}
				}));
		compiler.getTask(null, fileManager, diagnostics, null, null,
				compilationUnits).call();
		try {
			fileManager.close();
		} catch (IOException e) {
		}// not so important

		boolean success = true;
		// spracovanie chybovych hlaseni javac kompilatora a ich zapis do vyst. prudu out
		for (Diagnostic<?> diagnostic : diagnostics.getDiagnostics()) {
			if (diagnostic.getLineNumber() != -1) {
				String err = String.format("Error on line %d in %s%n",
						diagnostic.getLineNumber(),
						diagnostic.getMessage(Locale.ENGLISH));
				//log.severe("source: "+diagnostic.getSource());
				logger.error(new File(diagnostic.getSource().toString()));
				logger.error(err);

				success = false;
			} else {
				logger.info(diagnostic.getMessage(Locale.ENGLISH));
			}
		}
		if (success)
			logger.info("Compilation Successful");
	}

	/**
	 * Execute generated Java code of selected project as a new process.
	 * Output and error stream redirects to the graphic console, and
	 * input stream will be connected to the text field component.
	 */
	private void execute() {
		if (projectFolder() == null) {
			showError(NO_PROJECT_ERROR_MESSAGE);
			return;
		}
		console.clear();
		logger.debug("Starting program ...");
		logger.info("");// new line
		Runnable task = new Runnable() {
			public void run() {
				try {
					String command = String.format("java -cp %s:%s MainClass",
							classDir.getAbsolutePath(),
							libraryDir.getAbsolutePath());
					//logger.error(command);
					System.out.println(command);
					Process p = Runtime.getRuntime().exec(command);

					stdin = p.getOutputStream();
					BufferedReader output = new BufferedReader(
							new InputStreamReader(p.getInputStream()));
					BufferedReader err = new BufferedReader(
							new InputStreamReader(p.getErrorStream()));

					// must be read by single letters
					int c;
					while ((c = output.read()) != -1) {
						console.write("" + (char) c);
					}
					String line;
					//while ((line = output.readLine()) != null) {
					//    log.info(line);
					//}
					while ((line = err.readLine()) != null) {
						logger.error(line);
					}
					output.close();
					err.close();
					stdin.close();
					stdin = null;
					logger.debug("Program finished");
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		};
		new Thread(task).start();
	}

	/** This method is called from within the constructor to
	 * initialize the form.
	 * WARNING: Do NOT modify this code. The content of this method is
	 * always regenerated by the Form Editor.
	 */
	@SuppressWarnings("unchecked")
	//GEN-BEGIN:initComponents
	// <editor-fold defaultstate="collapsed" desc="Generated Code">
	private void initComponents() {

		graphViewer = new javax.swing.JDialog();
		jScrollPane3 = new javax.swing.JScrollPane();
		jPanel1 = new javax.swing.JPanel();
		input = new javax.swing.JTextField();
		consoleScrollPane = new javax.swing.JScrollPane();
		jComboBox1 = new javax.swing.JComboBox();
		jButton1 = new javax.swing.JButton();
		jScrollPane1 = new javax.swing.JScrollPane();
		jList1 = new javax.swing.JList();
		jButton2 = new javax.swing.JButton();
		compileToJava = new javax.swing.JButton();
		jButton3 = new javax.swing.JButton();
		showGraphButton = new javax.swing.JButton();
		jMenuBar1 = new javax.swing.JMenuBar();
		jMenu1 = new javax.swing.JMenu();
		jMenuItem1 = new javax.swing.JMenuItem();
		jMenuItem2 = new javax.swing.JMenuItem();
		jMenu2 = new javax.swing.JMenu();
		jMenuItem3 = new javax.swing.JMenuItem();
		jMenu3 = new javax.swing.JMenu();
		jMenuItem4 = new javax.swing.JMenuItem();
		jMenuItem5 = new javax.swing.JMenuItem();

		javax.swing.GroupLayout graphViewerLayout = new javax.swing.GroupLayout(
				graphViewer.getContentPane());
		graphViewer.getContentPane().setLayout(graphViewerLayout);
		graphViewerLayout
				.setHorizontalGroup(graphViewerLayout
						.createParallelGroup(
								javax.swing.GroupLayout.Alignment.LEADING)
						.addGroup(
								graphViewerLayout
										.createSequentialGroup()
										.addContainerGap()
										.addComponent(
												jScrollPane3,
												javax.swing.GroupLayout.DEFAULT_SIZE,
												376, Short.MAX_VALUE)
										.addContainerGap()));
		graphViewerLayout
				.setVerticalGroup(graphViewerLayout
						.createParallelGroup(
								javax.swing.GroupLayout.Alignment.LEADING)
						.addGroup(
								graphViewerLayout
										.createSequentialGroup()
										.addContainerGap()
										.addComponent(
												jScrollPane3,
												javax.swing.GroupLayout.DEFAULT_SIZE,
												276, Short.MAX_VALUE)
										.addContainerGap()));

		setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
		setTitle("BKPI");
		addWindowListener(new java.awt.event.WindowAdapter() {
			public void windowOpened(java.awt.event.WindowEvent evt) {
				formWindowOpened(evt);
			}
		});

		jPanel1.setBackground(new java.awt.Color(89, 89, 89));

		input.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent evt) {
				inputEntered(evt);
			}
		});

		javax.swing.GroupLayout jPanel1Layout = new javax.swing.GroupLayout(
				jPanel1);
		jPanel1.setLayout(jPanel1Layout);
		jPanel1Layout.setHorizontalGroup(jPanel1Layout
				.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
				.addComponent(input, javax.swing.GroupLayout.DEFAULT_SIZE, 453,
						Short.MAX_VALUE)
				.addComponent(consoleScrollPane,
						javax.swing.GroupLayout.DEFAULT_SIZE, 453,
						Short.MAX_VALUE));
		jPanel1Layout
				.setVerticalGroup(jPanel1Layout
						.createParallelGroup(
								javax.swing.GroupLayout.Alignment.LEADING)
						.addGroup(
								javax.swing.GroupLayout.Alignment.TRAILING,
								jPanel1Layout
										.createSequentialGroup()
										.addComponent(
												consoleScrollPane,
												javax.swing.GroupLayout.DEFAULT_SIZE,
												223, Short.MAX_VALUE)
										.addPreferredGap(
												javax.swing.LayoutStyle.ComponentPlacement.RELATED)
										.addComponent(
												input,
												javax.swing.GroupLayout.PREFERRED_SIZE,
												javax.swing.GroupLayout.DEFAULT_SIZE,
												javax.swing.GroupLayout.PREFERRED_SIZE)));

		jComboBox1.setModel(recentDirectories());
		jComboBox1.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent evt) {
				projectSelected(evt);
			}
		});

		jButton1.setText("Browse");
		jButton1.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent evt) {
				jButton1browsePerformed(evt);
			}
		});

		jList1.setBorder(new javax.swing.border.LineBorder(new java.awt.Color(
				89, 89, 89), 1, true));
		jList1.addMouseListener(new java.awt.event.MouseAdapter() {
			public void mouseClicked(java.awt.event.MouseEvent evt) {
				jList1fileMouseClicked(evt);
			}
		});
		jScrollPane1.setViewportView(jList1);

		jButton2.setText("Generate Code");
		jButton2.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent evt) {
				jButton2generatePerformed(evt);
			}
		});

		compileToJava.setText("Compile Code");
		compileToJava.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent evt) {
				compileToJavaActionPerformed(evt);
			}
		});

		jButton3.setText("Run");
		jButton3.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent evt) {
				jButton3execute(evt);
			}
		});

		showGraphButton.setText("Show Graph");
		showGraphButton.setToolTipText("Shows Project Structure");
		showGraphButton.setIconTextGap(1);
		showGraphButton.setRolloverEnabled(false);
		showGraphButton.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent evt) {
				showGraphButtonshowGraph(evt);
			}
		});

		jMenu1.setText("Excluded words");

		jMenuItem1.setText("Interactive manager ...");
		jMenuItem1.addMouseListener(new java.awt.event.MouseAdapter() {
			public void mousePressed(java.awt.event.MouseEvent evt) {
				jMenuItem1MousePressed(evt);
			}
		});
		jMenu1.add(jMenuItem1);

		jMenuItem2.setText("Interface manager ...");
		jMenuItem2.addMouseListener(new java.awt.event.MouseAdapter() {
			public void mousePressed(java.awt.event.MouseEvent evt) {
				jMenuItem2MousePressed(evt);
			}
		});
		jMenu1.add(jMenuItem2);

		jMenuBar1.add(jMenu1);

		jMenu2.setText("Compilation");

		jMenuItem3.setText("Compiled implementations manager ...");
		jMenuItem3.addMouseListener(new java.awt.event.MouseAdapter() {
			public void mousePressed(java.awt.event.MouseEvent evt) {
				jMenuItem3MousePressed(evt);
			}
		});
		jMenu2.add(jMenuItem3);

		jMenuBar1.add(jMenu2);

		jMenu3.setText("Debug");

		jMenuItem4.setText("Interface modification policy manager");
		jMenuItem4.addMouseListener(new java.awt.event.MouseAdapter() {
			public void mousePressed(java.awt.event.MouseEvent evt) {
				jMenuItem4MousePressed(evt);
			}
		});
		jMenu3.add(jMenuItem4);

		jMenuItem5.setText("Interface access policy manager");
		jMenuItem5.addMouseListener(new java.awt.event.MouseAdapter() {
			public void mousePressed(java.awt.event.MouseEvent evt) {
				jMenuItem5MousePressed(evt);
			}
		});
		jMenu3.add(jMenuItem5);

		jMenuBar1.add(jMenu3);

		setJMenuBar(jMenuBar1);

		javax.swing.GroupLayout layout = new javax.swing.GroupLayout(
				getContentPane());
		getContentPane().setLayout(layout);
		layout.setHorizontalGroup(layout
				.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
				.addGroup(
						layout.createSequentialGroup()
								.addContainerGap()
								.addGroup(
										layout.createParallelGroup(
												javax.swing.GroupLayout.Alignment.LEADING)
												.addGroup(
														layout.createSequentialGroup()
																.addComponent(
																		jComboBox1,
																		javax.swing.GroupLayout.PREFERRED_SIZE,
																		529,
																		javax.swing.GroupLayout.PREFERRED_SIZE)
																.addContainerGap())
												.addGroup(
														layout.createSequentialGroup()
																.addComponent(
																		jScrollPane1,
																		javax.swing.GroupLayout.PREFERRED_SIZE,
																		181,
																		javax.swing.GroupLayout.PREFERRED_SIZE)
																.addPreferredGap(
																		javax.swing.LayoutStyle.ComponentPlacement.RELATED)
																.addGroup(
																		layout.createParallelGroup(
																				javax.swing.GroupLayout.Alignment.LEADING)
																				.addComponent(
																						jPanel1,
																						javax.swing.GroupLayout.DEFAULT_SIZE,
																						javax.swing.GroupLayout.DEFAULT_SIZE,
																						Short.MAX_VALUE)
																				.addGroup(
																						layout.createSequentialGroup()
																								.addGroup(
																										layout.createParallelGroup(
																												javax.swing.GroupLayout.Alignment.TRAILING)
																												.addComponent(
																														jButton1,
																														javax.swing.GroupLayout.PREFERRED_SIZE,
																														88,
																														javax.swing.GroupLayout.PREFERRED_SIZE)
																												.addGroup(
																														javax.swing.GroupLayout.Alignment.LEADING,
																														layout.createSequentialGroup()
																																.addComponent(
																																		jButton2,
																																		javax.swing.GroupLayout.PREFERRED_SIZE,
																																		124,
																																		javax.swing.GroupLayout.PREFERRED_SIZE)
																																.addPreferredGap(
																																		javax.swing.LayoutStyle.ComponentPlacement.RELATED)
																																.addComponent(
																																		compileToJava,
																																		javax.swing.GroupLayout.PREFERRED_SIZE,
																																		119,
																																		javax.swing.GroupLayout.PREFERRED_SIZE)
																																.addPreferredGap(
																																		javax.swing.LayoutStyle.ComponentPlacement.RELATED)
																																.addComponent(
																																		jButton3,
																																		javax.swing.GroupLayout.PREFERRED_SIZE,
																																		66,
																																		javax.swing.GroupLayout.PREFERRED_SIZE)
																																.addPreferredGap(
																																		javax.swing.LayoutStyle.ComponentPlacement.RELATED)
																																.addComponent(
																																		showGraphButton,
																																		javax.swing.GroupLayout.PREFERRED_SIZE,
																																		111,
																																		javax.swing.GroupLayout.PREFERRED_SIZE)))
																								.addContainerGap()))))));
		layout.setVerticalGroup(layout
				.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
				.addGroup(
						javax.swing.GroupLayout.Alignment.TRAILING,
						layout.createSequentialGroup()
								.addContainerGap()
								.addGroup(
										layout.createParallelGroup(
												javax.swing.GroupLayout.Alignment.BASELINE)
												.addComponent(
														jComboBox1,
														javax.swing.GroupLayout.PREFERRED_SIZE,
														javax.swing.GroupLayout.DEFAULT_SIZE,
														javax.swing.GroupLayout.PREFERRED_SIZE)
												.addComponent(jButton1))
								.addPreferredGap(
										javax.swing.LayoutStyle.ComponentPlacement.RELATED)
								.addGroup(
										layout.createParallelGroup(
												javax.swing.GroupLayout.Alignment.TRAILING)
												.addGroup(
														layout.createSequentialGroup()
																.addGroup(
																		layout.createParallelGroup(
																				javax.swing.GroupLayout.Alignment.BASELINE)
																				.addComponent(
																						jButton2)
																				.addComponent(
																						compileToJava,
																						javax.swing.GroupLayout.DEFAULT_SIZE,
																						javax.swing.GroupLayout.DEFAULT_SIZE,
																						Short.MAX_VALUE)
																				.addComponent(
																						jButton3,
																						javax.swing.GroupLayout.DEFAULT_SIZE,
																						javax.swing.GroupLayout.DEFAULT_SIZE,
																						Short.MAX_VALUE)
																				.addComponent(
																						showGraphButton,
																						javax.swing.GroupLayout.DEFAULT_SIZE,
																						javax.swing.GroupLayout.DEFAULT_SIZE,
																						Short.MAX_VALUE))
																.addPreferredGap(
																		javax.swing.LayoutStyle.ComponentPlacement.RELATED)
																.addComponent(
																		jPanel1,
																		javax.swing.GroupLayout.DEFAULT_SIZE,
																		javax.swing.GroupLayout.DEFAULT_SIZE,
																		Short.MAX_VALUE))
												.addComponent(
														jScrollPane1,
														javax.swing.GroupLayout.DEFAULT_SIZE,
														284, Short.MAX_VALUE))));

		pack();
	}// </editor-fold>
	//GEN-END:initComponents

	private void jMenuItem5MousePressed(java.awt.event.MouseEvent evt) {
		IAPM.setVisible(true);
	}

	private void jMenuItem4MousePressed(java.awt.event.MouseEvent evt) {
		IMPM.setVisible(true);
	}

	private CompiledImplementationManager compiledImplementationsManager = new CompiledImplementationManager();

	private void jMenuItem3MousePressed(java.awt.event.MouseEvent evt) {
		ICIM.setVisible(true);
	}

	private void jMenuItem2MousePressed(java.awt.event.MouseEvent evt) {
		interface_manager_gui.setVisible(true);
	}

	private void jMenuItem1MousePressed(java.awt.event.MouseEvent evt) {
		interactive_manager_gui.setVisible(true);
	}

	InterfaceManagerGui interface_manager_gui;
	InteractiveManagerGui interactive_manager_gui;

	InteractiveCompiledImplementationsManager ICIM;

	ExcludedWordsManager EWM;

	private InterfaceModificationPolicyManager IMPM;
	private InterfaceAccessPolicyManager IAPM;

	private void formWindowOpened(java.awt.event.WindowEvent evt) {
		EWM = new ExcludedWordsManager();
		EWM.setImplementedLanguages(new ArrayList<String>(Arrays.asList("Java",
				"Csharp")));
		interface_manager_gui = new InterfaceManagerGui();
		interactive_manager_gui = new InteractiveManagerGui();
		interactive_manager_gui.SetExcludedWordsmanager(EWM);
		interface_manager_gui.SetExcludedWordsmanager(EWM);
		ICIM = new InteractiveCompiledImplementationsManager();
		ICIM.setCompiledImplementationsManager(compiledImplementationsManager);
		IMPM = new InterfaceModificationPolicyManager();
		IAPM = new InterfaceAccessPolicyManager();

	}

	private void inputEntered(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_inputEntered
		if (stdin != null) {
			//log.info(input.getText());
			String in = input.getText() + "\n";
			try {
				stdin.write(in.getBytes());
				stdin.flush();
				console.write(in);
			} catch (IOException ex) {
				logger.error("error in writing to program input stream");
			}
			input.setText("");
		}
	}//GEN-LAST:event_inputEntered

	private void projectSelected(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_projectSelected
		bProject = null;
		File project = projectFolder();
		if (project == null) {
			return;
		}
		// update recent projects list order
		recentProjects.remove(project.toString());
		recentProjects.add(0, project.toString());
		saveRecentProjectsList();

		if (project.exists()) {
			jList1.setModel(model(project));
			javaDir = new File(project, "java");
			classDir = new File(javaDir, "class");
		} else {
			logger.warn("Project " + project + " doesn't exist!");
		}
	}//GEN-LAST:event_projectSelected

	private void jButton1browsePerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jButton1browsePerformed
		JFileChooser fc = new JFileChooser();
		fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);

		int option = fc.showDialog(null, "select");
		if (option == JFileChooser.APPROVE_OPTION) {
			System.out.println(fc.getSelectedFile());
			recentProjects.add(0, fc.getSelectedFile().getAbsolutePath());
			jComboBox1.setModel(recentDirectories());
			saveRecentProjectsList();
			projectSelected(null);
		}
	}//GEN-LAST:event_jButton1browsePerformed

	private void jList1fileMouseClicked(java.awt.event.MouseEvent evt) {//GEN-FIRST:event_jList1fileMouseClicked
		if (evt.getClickCount() == 2) {
			Object selectedItem = jList1.getSelectedValue();
			if (selectedItem != null && selectedItem instanceof FileItem) {
				openFile(((FileItem) selectedItem).getFile().getAbsolutePath());
			}
		}
	}//GEN-LAST:event_jList1fileMouseClicked

	private void jButton2generatePerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jButton2generatePerformed
		parseProjectAsynchronously();
	}//GEN-LAST:event_jButton2generatePerformed

	private void compileToJavaActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_compileToJavaActionPerformed
		compileToJava();
	}//GEN-LAST:event_compileToJavaActionPerformed

	private void jButton3execute(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jButton3execute
		execute();
	}//GEN-LAST:event_jButton3execute

	private void showGraphButtonshowGraph(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_showGraphButtonshowGraph
		showGraphAsynchronously();
	}//GEN-LAST:event_showGraphButtonshowGraph

	public static void main(String args[]) {
		//BasicConfigurator.configure();
		//DOMConfigurator.configure("log4j.xml");
		PropertyConfigurator.configure("log4j.properties");

		try {
			UIManager
					.setLookAndFeel("com.sun.java.swing.plaf.nimbus.NimbusLookAndFeel");
		} catch (Exception ex) {
			logger.log(Level.ERROR, null, ex);
		}
		java.awt.EventQueue.invokeLater(new Runnable() {
			public void run() {
				new SimpleGui().setVisible(true);
			}
		});
	}

	//GEN-BEGIN:variables
	// Variables declaration - do not modify
	private javax.swing.JButton compileToJava;
	private javax.swing.JScrollPane consoleScrollPane;
	private javax.swing.JDialog graphViewer;
	private javax.swing.JTextField input;
	private javax.swing.JButton jButton1;
	private javax.swing.JButton jButton2;
	private javax.swing.JButton jButton3;
	private javax.swing.JComboBox jComboBox1;
	private javax.swing.JList jList1;
	private javax.swing.JMenu jMenu1;
	private javax.swing.JMenu jMenu2;
	private javax.swing.JMenu jMenu3;
	private javax.swing.JMenuBar jMenuBar1;
	private javax.swing.JMenuItem jMenuItem1;
	private javax.swing.JMenuItem jMenuItem2;
	private javax.swing.JMenuItem jMenuItem3;
	private javax.swing.JMenuItem jMenuItem4;
	private javax.swing.JMenuItem jMenuItem5;
	private javax.swing.JPanel jPanel1;
	private javax.swing.JScrollPane jScrollPane1;
	private javax.swing.JScrollPane jScrollPane3;
	private javax.swing.JButton showGraphButton;

	// End of variables declaration//GEN-END:variables

	static class FileItem {
		File file;

		public FileItem(File file) {
			this.file = file;
		}

		File getFile() {
			return file;
		}

		@Override
		public String toString() {
			return file.getName();
		}
	}
}
