package sk.tuke.bkpi.logging;

import java.io.FileInputStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.apache.log4j.AppenderSkeleton;
import org.apache.log4j.Level;
import org.apache.log4j.PropertyConfigurator;
import org.apache.log4j.spi.OptionHandler;

/**
 * Extends functionality of log4j library by allowing to define template
 * loggers/appenders in configuration file. Descendant loggers of any template
 * logger will have created and configured appenders (handlers) at runtime.
 * This class also provide loading of configuration from file. 
 * 
 * <p> For example, it is possible to redirect logging messages to the separated files
 * for loggers instances known only at runtime, but still to configure them in configuration
 * file.
 * 
 * @author Marcel Dancak
 */
public class Logger {
	
	private static Map<String, LoggerTemplate> loggerTemplates = new Hashtable<String, LoggerTemplate>();
	private static Map<String, AppenderTemplate> appenders;
	
	public static void configure(String filename) {
		PropertyConfigurator.configure(filename);
		
		loggerTemplates = new Hashtable<String, LoggerTemplate>();
		appenders = new Hashtable<String, AppenderTemplate>();
		
		Properties properties = new Properties();
		
		Pattern p = Pattern.compile("log4j\\.template\\.appender\\.(\\w+)((?:\\.\\w+)+)");
		
		//Pattern appenderProperty = Pattern.compile("log4j.template.appender.(\\w+).(\\w+)");
		Pattern appenderClass = Pattern.compile("log4j.template.appender.(\\w+)");
		try {
			properties.load(new FileInputStream(filename));
			for (Object k : properties.keySet()) {
				String key = k.toString();
				Matcher classMatcher = appenderClass.matcher(key);
				if (classMatcher.matches()) {
					//System.out.println("class: "+key);
					String appenderName = classMatcher.group(1);
					AppenderTemplate appender = new AppenderTemplate(Class.forName(properties.getProperty(key)));
					appenders.put(appenderName, appender);
				}
			}
			for (Object k : properties.keySet()) {
				String key = k.toString();
				Matcher mm = p.matcher(key);
				if (mm.matches()) {
					String appenderName = mm.group(1);
					String[] props = mm.group(2).substring(1).split("\\.");
					String value = properties.getProperty(key);
					//System.out.println("PROPERTY: "+key);
					//System.out.println("appender: "+appenderName);
					
					AppenderTemplate appender = appenders.get(appenderName);
					if (appender == null) {
						throw new NullPointerException("no appender with name: "+appenderName);
					}
					appender.addProperty(props, value);
					continue;
				}
				
				if (key.startsWith("log4j.template.category.")) {
					String basePackage = key.substring("log4j.template.category.".length());
					String value = properties.getProperty(key);
					
					List<AppenderTemplate> attachedAppenders = new ArrayList<AppenderTemplate>();
                    Level level = null;
                    boolean first = true;
					for (String v : value.split(",")) {
						if (appenders.containsKey(v.trim())) {
							attachedAppenders.add(appenders.get(v.trim()));
						} else if (first){
                            if (v.equals("INHERITED") || v.equals("NULL")) {
                                // ???
                            } else {
                                level = Level.toLevel(v);
                            }
                        }
                        first = false;
					}
					loggerTemplates.put(basePackage, new LoggerTemplate(attachedAppenders, level));
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Shorthand for getLogger(clazz.getName()). 
	 * @param clazz The class, which name will be used as logger name
	 * @return logger instance
	 */
	public static org.apache.log4j.Logger getLogger(Class<?> clazz) {
		//System.out.println("create logger: "+clazz.getName());
		return getLogger(clazz.getName());
	}
	
	/**
	 * Creates log4j logger instance with support for loggers templates functionality.
	 * @param name The name of the logger to retrieve.
	 * @return logger instance
	 */
	public static org.apache.log4j.Logger getLogger(String name) {
		org.apache.log4j.Logger logger = org.apache.log4j.Logger.getLogger(name);
		for (String registredName : loggerTemplates.keySet()) {
			if (name.startsWith(registredName)) {
				Map<String, String> parameters = new Hashtable<String, String>();
				parameters.put("{%ln}", name.substring(registredName.length())); //// {%ln} parameter (logger name)
				//System.out.println("create appender! "+parameters);
                
				for (AppenderTemplate appenderTemplate : loggerTemplates.get(registredName).appenders) {
					try {
						logger.addAppender(appenderTemplate.createAppender(name, parameters));
					} catch (Exception e) {
						e.printStackTrace();
					}
				}
                Level level = loggerTemplates.get(registredName).level;
                if (level != null) {
                    logger.setLevel(level);
                }
			}
		}

		return logger;
	}

	static class AppenderTemplate {
		private Class<?> appenderClass;
		private List<AppenderProperty> properties = new ArrayList<AppenderProperty>();
		
		public AppenderTemplate(Class<?> appenderClass) {
			this.appenderClass = appenderClass;
		}
		
		void addProperty(String[] propertiesNames, String value) {
			properties.add(new AppenderProperty(propertiesNames, value));
		}
		
		public AppenderSkeleton createAppender(String logger, Map<String, String> parameters) throws InstantiationException, IllegalAccessException, ClassNotFoundException, IllegalArgumentException, InvocationTargetException {
			Collections.sort(properties, new Comparator<AppenderProperty>() {
				@Override
				public int compare(AppenderProperty o1, AppenderProperty o2) {
					return o1.properties.length - o2.properties.length;
				}
				
			});
			//System.out.println("sorted: "+properties);
			
			AppenderSkeleton appender = (AppenderSkeleton) appenderClass.newInstance();
			for (AppenderProperty property : properties) {
				Object object;
				
				object = appender;// start always with appender object
				
				if (property.properties.length > 1) {
					propertiesLoop: for (int i = 0; i< property.properties.length-1; i++) {
						String propertyName = property.properties[i];
						String getter = String.format("get%s%s",
								Character.toUpperCase(propertyName.charAt(0)),
								propertyName.substring(1) );
						for (Method method : object.getClass().getMethods()) {
							if (method.getName().equals(getter) && method.getParameterTypes().length == 0) {
								try {
									object = method.invoke(object);
								} catch (Exception e) {
									e.printStackTrace();
								}
								continue propertiesLoop;
							}
						}
						throw new IllegalArgumentException(getter+" method  not found");
					}
				}
				
				String propertyName = property.properties[property.properties.length-1];
				String setter = String.format("set%s%s",
						Character.toUpperCase(propertyName.charAt(0)),
						propertyName.substring(1) );
				for (Method method : object.getClass().getMethods()) {
					if (method.getName().equals(setter) && method.getParameterTypes().length == 1) {
						Object value;
						Class<?> type = method.getParameterTypes()[0];
						if (type.isPrimitive()) {
							if (type == boolean.class) {
								value = Boolean.valueOf(property.value);
							} else if (type == int.class) {
								value = Integer.valueOf(property.value);
							} else {
								throw new RuntimeException(""+type);
							}
						} else {
							if (method.getParameterTypes()[0] == String.class) {
								String val = property.value;
								for (String key: parameters.keySet()) {
									val = val.replace(key, parameters.get(key));
								}
								value = val;
							} else {
								value = Class.forName(property.value).newInstance();
							}
						}
						method.invoke(object, value);
					}
				}
				//throw new IllegalArgumentException(setter+" method  not found in class: "+object.getClass());
			}
			if (appender instanceof OptionHandler) {
				((OptionHandler) appender).activateOptions();
			}
			return appender;
		}
	}
	
	static class AppenderProperty {
		String[] properties;
		String value;
		
		public AppenderProperty(String[] property, String value) {
			this.properties = property;
			this.value = value;
		}
		
		@Override
		public String toString() {
			return Arrays.asList(properties).toString();
		}
	}
	
    private static class LoggerTemplate {
        private List<AppenderTemplate> appenders;
        private Level level;

        public LoggerTemplate(List<AppenderTemplate> appenders, Level level) {
            this.appenders = appenders;
            this.level = level;
        }
    }
}
