package sk.tuke.bkpi.model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.TreeSet;

import org.apache.log4j.Logger;

import sk.tuke.bkpi.generator.CompiledImplementationsResolver;
import sk.tuke.bkpi.model.clause.Import;
import sk.tuke.bkpi.model.clause.SeenMachine;
import sk.tuke.bkpi.model.component.AbstractComponent;
import sk.tuke.bkpi.model.component.Implementation;
import sk.tuke.bkpi.model.component.Machine;
import sk.tuke.bkpi.model.component.Refinement;

/**
 * B-Method project, collecting all of its modules. Input is a list of components
 * and the result is a list of modules with dependencies between them. Project
 * also contains imports tree graph.
 * 
 * @author Marcel Dancak
 */
public class Project {

	private List<Implementation> implementations;
	private List<Refinement> refinements;
	private List<Machine> machines;
	
	private Map<Machine, Module> modules;
	private List<Module> mainModules = new ArrayList<Module>();
	private Map<Module, Node<Module>> mainModulesGraphs = new Hashtable<Module, Node<Module>>();
	
	private static Logger log = sk.tuke.bkpi.logging.Logger.getLogger(Project.class);
	/*
	public Project(List<Machine> machines, List<Refinement> refinements, List<Implementation> implementations) {
		this.machines = machines;
		this.refinements = refinements;
		this.implementations = implementations;
		
		CIR.createImplementationsPerMachine(machines, implementations);
		
		createModules();
	}*/
	
	public Project(List<Machine> machines, List<Refinement> refinements, List<Implementation> implementations, Hashtable <String, String> MIPairs) {
		this.machines = machines;
		this.refinements = refinements;
		this.implementations = implementations;
		
		CIR.createImplementationsPerMachine(machines, implementations);
		CIR.setMIPairs(MIPairs);
		
		createModules();
	}
	
	private CompiledImplementationsResolver CIR = new CompiledImplementationsResolver();
	
	private final void createModules() {
		
		Set<String> imported = new TreeSet<String>();
		List<Machine> withoutRefinement = new ArrayList<Machine>(machines);
		
		modules = new Hashtable<Machine, Module>();
		List<Refinement> unattachedRefinements = new ArrayList<Refinement>(refinements);
		
		for (Implementation impl : implementations) {
			if (CIR.generateImplementation(impl.abstractMachine(), impl))
			{
			List<Refinement> ref = new ArrayList<Refinement>();
			AbstractComponent refinedComponent = impl.refinedComponent();
			while (refinedComponent instanceof Refinement) {
				ref.add(0, (Refinement) refinedComponent);
				refinedComponent = ((Refinement)refinedComponent).refinedComponent();
			} 
			Module module = new Module(impl.abstractMachine(), ref, impl);
			modules.put(impl.abstractMachine(), module);
			unattachedRefinements.removeAll(ref);
			withoutRefinement.remove(impl.abstractMachine());
			for (Import imp : impl.imports()) {
				imported.add(imp.machine().name());
			}
			}//
		}
		int i = 0;
		while(i < unattachedRefinements.size()) { //unattachedRefinements.size() > 0 &&
		//for (Refinement refinement : unattachedRefinements) {
			Refinement refinement = unattachedRefinements.get(i);
			i++;
			boolean isLastRefinement = true;
			for (Refinement r : unattachedRefinements) {
				if (r.refinedComponent() == refinement) {
					isLastRefinement = false;
					break;
				}
			}
			if (isLastRefinement) {
				List<Refinement> refs = new ArrayList<Refinement>();
				refs.add(0, refinement);
				AbstractComponent refinedComponent = refinement.refinedComponent();
				while (refinedComponent instanceof Refinement) {
					refs.add(0, (Refinement) refinedComponent);
					refinedComponent = ((Refinement)refinedComponent).refinedComponent();
				}
				
				Machine abstractMachine = (Machine) refinedComponent;
				if (modules.containsKey(abstractMachine)) {
					throw new RuntimeException("more independent refinements of abstract machine: "+abstractMachine);
				}
				modules.put(abstractMachine, new Module((Machine) refinedComponent, refs, null));
				unattachedRefinements.removeAll(refs);
				withoutRefinement.remove(abstractMachine);
				i = 0;
			}
		}
		if (unattachedRefinements.size() > 0) {
			throw new RuntimeException("Some refinements left in pool!");
		}
		
		for (Machine machine : withoutRefinement) {
			Module module = new Module(machine);
			assert !modules.containsKey(machine);
			modules.put(machine, module);
		}
		
		for (Module module : modules.values()) {
			if (module.implementation() != null) {
				List<Module> importedModules = new ArrayList<Module>();
				for (Import imp : module.implementation().imports()) {
					importedModules.add(modules.get(imp.machine()));
				}
				module.setImportedModules(importedModules);
			}
		}
		
		for (Machine machine : withoutRefinement) {
			//mainModulesGraphs.put(modules.get(machine), moduleTree(modules.get(machine)));
		}
		
		// find roots (main module(s)) in project imports tree
		for (Implementation impl : implementations) {
			if (!imported.contains(impl.abstractMachine().name())) {
				Module rootModule = modules.get(impl.abstractMachine());
				Node<Module> graph = moduleTree(rootModule);
				mainModules.add(rootModule);
				mainModulesGraphs.put(rootModule, graph);
			}
		}
		
		log.debug("Machines without implementation: "+withoutRefinement);
		log.debug("IMPORTED MACHINES: "+imported);
	}
	
	public List<Machine> machines() {
		return Collections.unmodifiableList(machines);
	}

	public List<Implementation> implementations() {
		return Collections.unmodifiableList(implementations);
	}
	
	public Collection<Module> modules() {
		return modules.values();
	}
	
	public List<Module> rootModules() {
		return mainModules;
	}
	
	public Module moduleOfMachine(Machine machine) {
		return modules.get(machine);
	}

	public Map<Module, Node<Module>> rootModulesGraphs() {
		return mainModulesGraphs;
	}

	
	/*
	 * Graph building
	 */
	
	private Node<Module> moduleTree(Module module) {
		Node<Module> node = new Node<Module>();
		node.module = module;
		node.name = module.machine().name();
		// build import's tree
		importsChains(node);
		
		// and now sees chains
		seesChains(node);
		return node;
	}
	
	private void seesChains(Node<Module> node) {
		log.info("Process Sees Chains: "+node);
		
		List<Edge<Module>> seenNodes = new ArrayList<Edge<Module>>();
		//if (node.module.implementation() != null) {
		for (SeenMachine seen : node.module.lastComponent().seenMachines()) {
			log.info("SEARCH: "+seen.path()+" "+seen.machine().name());
			
			Node<Module> seenNode = null;
			// check every ancestor till ...
			Edge<Module> e;
			Node<Module> ancestor = node;
			List<Edge<Module>> source = new ArrayList<Edge<Module>>();
			List<ImportEdge<Module>> target = null;
			findAncesor: while((e = ancestor.parentEdge)!=null) {
				
				Node<Module> checkingAncestor = e.parent;
				source.add(0, e);
				log.info("check acnestor: "+checkingAncestor);
				test: for (Node<Module> n : checkingAncestor.allChildren()) {
					if (n.module.machine() == seen.machine()) {
						log.info("\tcheck "+n);
						ImportEdge<Module> e2;
						Node<Module> p2 = n;
						Stack<String> reversedPath = new Stack<String>();
						target = new ArrayList<ImportEdge<Module>>();
						while((e2 = p2.parentEdge) != null && p2 != checkingAncestor) {
							if (e2.name != null && e2.name.length() > 0) {
								reversedPath.push(e2.name);
							}
							target.add(0, e2);
							p2 = e2.parent;
						}
						log.info("\treversed: "+reversedPath);
						
						// compare sees paths
						log.info("\tcompare: "+seen.path()+ " and "+reversedPath);
						if (seen.path().size() == reversedPath.size()) {
							for (String p : seen.path()) {
								log.info("checking: "+p +" equals "+reversedPath.peek()+" ?: "+p.equals(reversedPath.peek()));
								if (!p.equals(reversedPath.pop())) {
									continue test;
								}
							}
							log.info("Yeah "+seen+ " is correct");
							seenNode = n;
							break findAncesor;
						}
					}
				}
				
				checkingAncestor.module.addSeenReferenceDelegation(seen);
				ancestor = e.parent;
			}
			if (seenNode == null) {
				if (seen.path().size() == 0) {
					//modules.get(seen.machine())
					for (Node<Module> n : mainModulesGraphs.values()) {
						
						for (Node<Module> ch : n.allChildren()) {
							if (ch.module.machine() == seen.machine()) {
								seenNode = ch;
							}
						}
					}
					seenNode = new Node<Module>();
					seenNode.module = moduleOfMachine(seen.machine());
					seenNode.name = seen.machine().name();
					seen.setIsImported(false);
					if (seenNode == null) {
						throw new IllegalStateException("Invalid sees chain: "+node.module.lastComponent() +" -> "+seen.fullName());
						//continue;
					}
					
					System.out.println(node+" |-> "+seen);
					seenNodes.add(new Edge<Module>(node, seenNode, "static"));
                    log.warn(seen.machine().name() +" is seen, but not imported");
					//throw new IllegalStateException("Invalid sees chain: "+node.module.lastComponent() +" -> "+seen.fullName());
					continue;
				} else {
					// collect components names with this not correct sees chain
					List<String> corruptedComponents = new ArrayList<String>();
					for (SeenMachine seenMachine : node.module.implementation().seenMachines()) {
						if (seenMachine.equals(seen)) {
							corruptedComponents.add(node.module.implementation().name());
						}
					}
					String msg = String.format("Sees chain: %s defined in: %s is not valid (not imported in project)", seen.fullName(), corruptedComponents);
					throw new IllegalStateException(msg);
				}
			}
			seen.setIsImported(true);
			//s.parent.imported.indexOf(s);
			Node<Module> seenRoot = source.get(0).parent;
			System.out.println("root: "+seenRoot);
			System.out.println("source path: "+source);
			System.out.println("target path: "+target);
			
			List<Import> importedPath = new ArrayList<Import>(target.size());
			for (ImportEdge<Module> edge : target) {
				importedPath.add(edge.importedMachine);
			}
			seenRoot.module.addSeesDependency(source.get(0).child.module.machine(), target.get(0).child.module.machine());
			seenRoot.module.addImportPath(seen, importedPath);
			seenNodes.add(new Edge<Module>(node, seenNode, seen.prefixString()));
			
		}
		//}
		node.seen = seenNodes;
		for (Edge<Module> e : node.imported) {
			seesChains(e.child);
		}
	}
	
	private void importsChains(Node<Module> node) {
		List<ImportEdge<Module>> nodes = new ArrayList<ImportEdge<Module>>();
		
		if (node.module.implementation() != null) { // skip basic library modules
			List<Import> compositions = new ArrayList<Import>(node.module.implementation().imports());
			compositions.addAll(node.module.implementation().extendedMachines());
			for (Import imp : compositions) {
				Node<Module> n = new Node<Module>();
				ImportEdge<Module> edge = new ImportEdge<Module>(node, n, imp);
				n.module = modules.get(imp.machine());
				n.parentEdge = edge;//new Edge(node, imp.instanceName());
				
				// build name by traversing imports tree towards root
				List<Machine> acnestors = new ArrayList<Machine>();
				Stack<String> path = new Stack<String>();
				Node<Module> x = n;
				Edge<Module> e;
				while((e = x.parentEdge) != null) {
					if (e.name != null && e.name.length() > 0) {
						path.push(e.name);
					}
					acnestors.add(e.parent.module.machine());
					x = e.parent;
				}
				
				String prefix = "";
				while(!path.isEmpty()) {
					prefix += path.pop()+".";
				}
				n.name = prefix+imp.machine().name();
				if (n.name.startsWith("."))
					throw new IllegalStateException("start with .");
				if (node.isAcnestor(n)) {
					//throw new RuntimeException("imports cycle: "+node.module.implementation() + "->" + imp + " cycle with "+n.module);
				}
				System.out.println(n.module);
				System.out.println("ATTENTION: " + n.module.machine());
				if (acnestors.contains(n.module.machine())) {
					//throw new RuntimeException("imports cycle: ");
					node.invalidImports.add(edge);
				} else {
					importsChains(n);
					nodes.add(edge); //new Edge(n, imp.instanceName())
				}
			}
		}
		node.imported = nodes;
	}
	
	public static class ImportEdge<T> extends Edge<T> {
		Import importedMachine;
		public ImportEdge(Node<T> parent, Node<T> child, Import importedMachine) {
			super(parent, child, importedMachine.prefixString());
			this.importedMachine = importedMachine;
		}
	}
	
	public static class Edge<T> {
		String name;
		Node<T> parent;
		Node<T> child;
		//Node targetNode;
		
		public Edge(Node<T> parent, Node<T> child, String name) {
			this.parent = parent;
			this.child = child;
			this.name = name;
		}
		
		public String getName() {
			return name;
		}
		
		public Node<T> child() {
			return child;
		}
		
		@Override
		public String toString() {
			String x = parent+"->"+child;
			return name!=null? x+"("+name+")" : x;
		}
	}
	
	public static class Node<T> {
		String name;
		T module;
		public List<ImportEdge<T>> imported;
		List<ImportEdge<T>> invalidImports = new ArrayList<ImportEdge<T>>();
		
		List<Edge<T>> seen;
		ImportEdge<T> parentEdge;
		boolean processed;
		
		public List<Edge<T>> seen() {
			return seen;
		}
		public String getName() {
			return name;
		}
		public T getModule() {
			return module;
		}
		
		public List<ImportEdge<T>> imported() {
			return imported;
		}
		
		public List<Node<T>> allChildren() {
			List<Node<T>> children = new ArrayList<Node<T>>();
			children(this, children);
			return children;
		}
		
		void children(Node<T> parent, List<Node<T>> list) {
			list.add(parent);
			for (Edge<T> imported : parent.imported) {
				children(imported.child, list);
			}
		}
		
		boolean isAcnestor(Node<T> node) {
			Node<T> acnestor = node;
			Edge<T> e;
			while((e = acnestor.parentEdge) != null) {
				
				if (acnestor == node) {
					return true;
				}
				acnestor = e.parent;
			}
			return false;
		}
		
		@Override
		public String toString() {
			return name;
		}
	}
}
