package sk.tuke.bkpi.parser;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Stack;
import org.apache.log4j.Logger;

import sk.tuke.bkpi.model.component.Component;
import sk.tuke.bkpi.model.component.Machine;
import sk.tuke.bkpi.model.data.ConcreteConstant;
import sk.tuke.bkpi.model.data.ConcreteVariable;
import sk.tuke.bkpi.model.data.ForeignData;
import sk.tuke.bkpi.model.data.ForeignVariable;
import sk.tuke.bkpi.model.data.LocalVariable;
import sk.tuke.bkpi.model.data.OperationInputParameter;
import sk.tuke.bkpi.model.data.OperationOutputParameter;
import sk.tuke.bkpi.model.clause.ComponentReference;
import sk.tuke.bkpi.model.clause.EnumeratedSet;
import sk.tuke.bkpi.model.data.DataField;
import sk.tuke.bkpi.model.data.ForeignData.Chain;
import sk.tuke.bkpi.model.types.Type;

/**
 * Manages data of B-Method components during parsing. Supports multiple visibility
 * scopes.
 * 
 * @author Marcel Dancak
 */
public class DataContext {
	
	private static Logger defaultLogger = sk.tuke.bkpi.logging.Logger.getLogger(DataContext.class);
	
	private Logger logger;
	private Map<String, DataField> machineDataTable = new Hashtable<String, DataField>();
	private Map<String, Stack<DataField>> localDataTable = new Hashtable<String, Stack<DataField>>();
	
	public DataContext() {
		logger = defaultLogger;
	}
	
	/**
	 * Creates named <code>DataContext</code>, that can be easier to debug.
	 * @param name sufix of the class logger
	 */
	public DataContext(String name) {
		this.logger = sk.tuke.bkpi.logging.Logger.getLogger(DataContext.class.getName()+"."+name);
	}
	
	public Logger logger() {
		return logger;
	}
	
	public void addDataItems(ComponentReference machine, List<? extends DataField> items) {
		logger.info("add multiple data: "+items + "with prefix: "+machine.prefixString());
		for (DataField data : items)
			addDataItem(machine, data);
	}
	
	public void addDataItem(ComponentReference machine, DataField data) {
		if (data instanceof EnumeratedSet) {
			for (EnumeratedSet.Item item : ((EnumeratedSet) data).items()) {
				addDataItem(machine, item);  
			}
		}
		addDataItem(new ForeignVariable(machine, data), machine.prefixString());
	}
	
	public void addDataItems(List<? extends DataField> items) {
		logger.info("add multiple data: "+items);
		for (DataField data : items)
			addDataItem(data);
	}
	
	public Collection<DataField> allMachineData() {
		return machineDataTable.values();
	}
	
	private Map<Type, Type> typeSubstitutions = new Hashtable<Type, Type>();
	public void addTypeSubstitution(Type T1, Type T2) {
		typeSubstitutions.put(T1, T2);
	}
	
	private DataField afterSubstitution(DataField data) {
		if (!data.isTyped())
			return data;
		DataField old = data;
		//if (true) return data;
		//if (data.name().equals("ee")) return data;
		for (Type T1 : typeSubstitutions.keySet()) {
			if (data.type().contains(T1)) {
				logger.info("\tcontains: "+T1);
				Type newType = data.type().replace(T1, typeSubstitutions.get(T1));
				logger.info("after replace by "+ typeSubstitutions.get(T1) +" : "+newType);
				logger.info(""+data.getClass());
				DataField nData = data.newData();
				logger.info("\t1: "+data.type());
				nData.typeData(newType);
				logger.info("\t2: "+data.type());
				data = nData;
				logger.info("\tresult: "+data.type());
			}
		}
		if (old != data) {
			logger.info(""+typeSubstitutions);
			logger.info(old + " replaced by "+data);
			logger.info(old.getClass() + " replaced by "+data.getClass());
			//return old;
		}
		return data;
	}
	
	public void addDataItem(DataField data) {
		addDataItem(data, null);
	}
	
	public void addDataItem(DataField data, String prefix) {
		if (data instanceof EnumeratedSet) {
			for (EnumeratedSet.Item item : ((EnumeratedSet) data).items()) {
				addDataItem(item);  
			}
		}
		String name = prefix != null && prefix.length() > 0? prefix+"."+data.name() : data.name();
		
		if (data instanceof LocalVariable || data instanceof OperationOutputParameter || data instanceof OperationInputParameter) {
			logger.info("adding local data: "+name);
			if (!localDataTable.containsKey(name)) {
				Stack<DataField> dataList = new Stack<DataField>();
				dataList.push(afterSubstitution(data));
				localDataTable.put(name, dataList);
			} else {
				localDataTable.get(name).push(data);
			}
		} else {
			if (!machineDataTable.containsKey(name)) {
				logger.info("adding machine data: "+name+ " type: "+data.type()+ " class: "+data.getClass());
				machineDataTable.put(name, afterSubstitution(data));
			} else {
				DataField oldData = machineDataTable.get(name);
				//if (oldData instanceof RefinedVariable && data instanceof ConcreteVariable) {
					machineDataTable.remove(name);
					machineDataTable.put(name, data);
				//} else {
				//	throw new RuntimeException("Data Context already contains machine data: "+oldData +" "+oldData.getClass());
				//}
			}
		}
	}

	public boolean hasData(String name) {
		return machineDataTable.containsKey(name) || localDataTable.containsKey(name);
	}

	public DataField getData(String name) {
		if (localDataTable.containsKey(name))
			return localDataTable.get(name).peek();
		if (machineDataTable.containsKey(name))
			return machineDataTable.get(name);
		
		return null;
	}

	public void removeData(String name) {
		logger.info("Remove data: "+name);
		if (machineDataTable.containsKey(name)) {
			machineDataTable.remove(name);
		}
		if (localDataTable.containsKey(name)) {
			localDataTable.get(name).pop();
			if (localDataTable.get(name).size() == 0) {
				localDataTable.remove(name);
			}
		}
	}

	public void removeDataItems(List<? extends DataField> items) {
		for (DataField data : items)
			removeData(data.name());
	}
	
	public List<DataField> getAllData() {
		List<DataField> list = new ArrayList<DataField>(machineDataTable.values());
		for (Stack<DataField> local : localDataTable.values()) {
			list.add(local.peek());
		}
		return list;
	}
	
	@SuppressWarnings("unchecked")
	public <T extends DataField> List<T> dataOf(Class<T> type) {
		List<T> list = new ArrayList<T>();
		for (String dataName : localDataTable.keySet()) {
			DataField data = getData(dataName);
			if (data.getClass().equals(type)) {
			//if (type.isInstance(data)) {
				list.add((T)data);
			}
		}
		for (DataField data : machineDataTable.values()) {
			if (data.getClass().equals(type))
				list.add((T)data);
		}
		return list;
	}
	
	public void addMachineConstantsAndSets(Machine machine, Chain chain) {
		//String log = String.format("ADD %s data of machine: %s%s", chain, prefix!=null? prefix+"." : "", reference.machine().name());
		//logger.info(log);
		addDataItems(machine, chain, machine.concreteConstants());
		addDataItems(machine, chain, machine.abstractConstants());
		addDataItems(machine, chain, machine.sets());
		addDataItems(machine, chain, machine.includedSets());
	}
	
	public void addDataItems(Machine machine, Chain chain, List<? extends DataField> items) {
		for (DataField data : items) {
			if (data instanceof EnumeratedSet) {
				for (EnumeratedSet.Item item : ((EnumeratedSet) data).items()) {
					addDataItem(new ForeignData(machine, item));  
				}
			}
			addDataItem(new ForeignData(machine, data));
		}
	}
	
	public void addMachineVariables(ComponentReference reference) {
		addDataItems(reference, reference.machine().concreteVariables());
		addDataItems(reference, reference.machine().abstractVariables());
	}
	/*
	public void addRefinementData(Refinement refinement) {
		addMachineData(refinement);
		AbstractComponent refinedComponent = refinement.refinedComponent();
		while(refinedComponent instanceof Refinement) {
			addDataItems(refinedComponent.concreteConstants());
			addDataItems(refinedComponent.concreteVariables());
			refinedComponent = ((Refinement) refinedComponent).refinedComponent();
		}
		addDataItems(refinedComponent.concreteConstants());
		addDataItems(refinedComponent.concreteVariables());
	}
	*/
	private void addComponentData(Component component) {
		for (ConcreteConstant constant : component.concreteConstants()) {
			addDataItem(constant);
		}
		for (ConcreteVariable variable : component.concreteVariables()) {
			addDataItem(variable);
		}
	}
	
	@Override
	public String toString() {
		return "Data Context: [" + machineDataTable +" Local variables:" +localDataTable+ "]";
	}
}
