package sk.tuke.bkpi.parser;

import java.io.File;
import java.util.Collection;
import java.util.Collections;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

/**
 * Encapsulates mapping of B-Method language definitions by thier names.
 * @author Marcel Dancak
 */
public class DefinitionsContext {

	private Map<String, Definition> definitions = new Hashtable<String, Definition>();
	
	public void addDefinition(Definition definition) {
		if (definitions.containsKey(definition.name))
			throw new IllegalArgumentException("definition already defined: "+definition.name);
		
		definitions.put(definition.name, definition);
	}
	
	public boolean hasDefinition(String name) {
		return definitions.containsKey(name);
	}
	
	public Definition getDefinition(String name) {
		return definitions.get(name);
	}
	
	public Collection<Definition> listDefinitions() {
		return definitions.values();
	}
	
	@Override
	public String toString() {
		StringBuilder names = new StringBuilder();
		for (String name: definitions.keySet()) {
			names.append(name+", ");
		}
		return "Definition Context["+names+"]";
	}
	
	public static class Definition {
		
		public final String name;
		public final List<String> parameters;
		public final String text;
		
		public final File source;
		public final int beginLine;
		public final int beginColumn;
		
		public Definition(String name, List<String> parameters, String text) {
			this(name, parameters, text, null, 0, 0);
		}
		
		public Definition(String name, List<String> parameters, String text, File source, int row, int column) {
			this.name   = name;
			this.parameters = parameters != null? Collections.unmodifiableList(parameters) : Collections.<String>emptyList();
			this.text   = text;
			this.source = source;
			this.beginLine    = row;
			this.beginColumn = column;
		}
		/**
		 * Returns definition's text with substituted arguments  
		 * @param arguments definition's arguments
		 * @return definition's text
		 */
		public String getText(List<String> arguments) {
			assert arguments.size() == parameters.size();
			//String mergedText = text;
			StringBuilder sb = new StringBuilder(text);
			for (int i = 0; i < arguments.size(); i++) {
				String parameter = parameters.get(i);
				
				int pos = 0;
				while((pos = sb.indexOf(parameter, pos)) != -1) {
					if ((pos == 0 || !Character.isLetter(sb.charAt(pos-1))) &&
							(pos+parameter.length() >= sb.length() || !Character.isLetter(sb.charAt(pos+parameter.length())))) {
						sb.replace(pos, pos+parameter.length(), arguments.get(i));
						pos += arguments.get(i).length();
					} else {
						//System.out.println("before: "+sb.charAt(pos-1));
						//System.out.println("after: "+sb.charAt(pos+parameter.length()+1));
						pos += parameter.length();
					}
				}
				//mergedText = mergedText.replace(parameter, arguments.get(i));
			}
			//if (true) throw new RuntimeException(sb+" vs "+mergedText);
			//if (!mergedText.equals(sb.toString())) throw new RuntimeException(sb+" vs "+mergedText);
			//return mergedText;
			return sb.toString();
		}
		
		@Override
		public String toString() {
			return String.format("Definition[%s %s, file:%s position[%d, %d] text:%s]", name, parameters, source, beginLine, beginColumn, text);
		}
	}
}
