package sk.tuke.bkpi.parser;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import org.apache.log4j.Logger;

import sk.tuke.bkpi.model.Project;
import sk.tuke.bkpi.model.component.Component;
import sk.tuke.bkpi.model.component.Implementation;
import sk.tuke.bkpi.model.component.Machine;
import sk.tuke.bkpi.model.component.Refinement;
import sk.tuke.bkpi.parser.javacc.Parser;

/**
 * 
 * Encapsulates process of parsing a whole B-Method project files.
 * Result from parsing is data model of class <code>Project</code>.
 * 
 * @author Marcel Dancak
 */
public class ProjectParser implements DependencyResolver {
	
	private static Logger logger = sk.tuke.bkpi.logging.Logger.getLogger(ProjectParser.class);
	
	public static String MACHINE_EXTENSION        = ".mch";
	public static String REFINEMENT_EXTENSION     = ".ref";
	public static String IMPLEMENTATION_EXTENSION = ".imp";
	
	private List<Implementation> implementations;
	private List<Refinement> refinements;
	private List<Machine> machines;
	
	private File projectDirectory;

	private Map<String, File> projectFiles = new Hashtable<String, File>();
	private Map<String, File> libraryFiles = new Hashtable<String, File>();
	private Collection<String> librariesPaths = Collections.emptyList();
	
	private Map<String, Component> processedComponents;
	// to keep track of parsing dependency files
	private Stack<File> lastFile = new Stack<File>();
	
	/**
	 * Creates B-Method project parser.
	 * @param path The B-Method project folder
	 */
	public ProjectParser(File path) {
		projectDirectory = path;
		projectFiles = collectBFiles(projectDirectory);
	}
	
	public void setLibraries(String ... libraries) {
		setLibraries(Arrays.asList(libraries));
	}
	
	public void setLibraries(Collection<String> libraries) {
		librariesPaths = libraries;
		libraryFiles = new Hashtable<String, File>();
		
		for (String libraryPath : librariesPaths) {
			File library = new File(libraryPath);
			if (!library.exists()) {
				//throw new FileNotFoundException();
				throw new ProjectParseException("Library directory does not exist: "+library, library);
			}
			if (!library.isDirectory()) {
				throw new ProjectParseException("Library is not directory", library);
			}
			libraryFiles.putAll(collectBFiles(library));
		}
		//projectFiles.putAll(libraryFiles);
	}
	
	private Map<String, File> collectBFiles(File directory) {
		Map<String, File> bFiles = new Hashtable<String, File>();
		for (File f : directory.listFiles()) {
			if (f.isFile()) {
				String fileName = f.getName();
				if (fileName.endsWith(MACHINE_EXTENSION))
					bFiles.put(fileName.substring(0, fileName.lastIndexOf(MACHINE_EXTENSION)), f.getAbsoluteFile());
				else if (fileName.endsWith(REFINEMENT_EXTENSION))
					bFiles.put(fileName.substring(0, fileName.lastIndexOf(REFINEMENT_EXTENSION)), f.getAbsoluteFile());
				else if (fileName.endsWith(IMPLEMENTATION_EXTENSION))
					bFiles.put(fileName.substring(0, fileName.lastIndexOf(IMPLEMENTATION_EXTENSION)), f.getAbsoluteFile());
			}
		}
		return bFiles;
	}
	
	/**
	 * Parse whole project of B-Method and produce it's data model.
	 * 
	 * @return Data model of the B-Method project
	 * @throws ProjectParseException
	 * @throws FileNotFoundException
	 *//*
	public Project parseProject() throws ProjectParseException, FileNotFoundException {
		implementations = new ArrayList<Implementation>();
		machines = new ArrayList<Machine>();
		refinements = new ArrayList<Refinement>();
		
		processedComponents = new Hashtable<String, Component>();
		
		for (String component : projectFiles.keySet()) {
			Component c = getComponent(component);
		}
		return new Project(machines, refinements, implementations);
	}*/
	
	public Project parseProject(Hashtable <String, String> MIPairs) throws ProjectParseException, FileNotFoundException {
		implementations = new ArrayList<Implementation>();
		machines = new ArrayList<Machine>();
		refinements = new ArrayList<Refinement>();
		
		processedComponents = new Hashtable<String, Component>();
		
		for (String component : projectFiles.keySet()) {
			Component c = getComponent(component);
		}
		return new Project(machines, refinements, implementations, MIPairs );
	}
	
	/**
	 * Returns data model of component with name <code>componentName</code>.
	 * If component wasn't already parser, it will be. Otherwise it will
	 * be returned from cache.
	 * 
	 * @param componentName The name of the component
	 * @return Data model of the component
	 */
	@Override
	public Component getComponent(String componentName) {
		
		if (processedComponents.containsKey(componentName)) {
			return processedComponents.get(componentName);
		}
		
		File fileToParse;
		if (projectFiles.containsKey(componentName)) {
			fileToParse = projectFiles.get(componentName);
		} else if (libraryFiles.containsKey(componentName)) {
			fileToParse = libraryFiles.get(componentName);
		} else {
			//throw new ProjectParseException("Missing component: "+componentName);
			throw new RuntimeException("Missing component: "+componentName);
		}
		
		logger.info("## Parsing: "+componentName);
		lastFile.push(fileToParse);
		Parser parser = null;
		try {
			parser = Parser.myParser(fileToParse, this);
		} catch (FileNotFoundException e1) {
			e1.printStackTrace();
		}
		
		Component c;
		try {
			PreprocessData preprocessData = parser.Preprocess();
			parser.ReInit(new FileInputStream(fileToParse));
			parser.setPreprocessedData(preprocessData);
			c = parser.Component();
			processedComponents.put(componentName, c);// check with c.name()
		} catch (Exception e) {
			throw new ProjectParseException(e, lastFile.peek());
		}
		lastFile.pop();
		logger.info("## End of Parsing: "+componentName);
		if (c instanceof Implementation)
			implementations.add((Implementation) c);
		else if (c instanceof Machine)
			machines.add((Machine) c);
		else if (c instanceof Refinement)
			refinements.add((Refinement) c);
		return c;
	}
}
